<?php

namespace Wpo\Sync;

// Prevent public access to this script
defined('ABSPATH') or die();

use \WP_Error;
use \Wpo\Core\Permissions_Helpers;
use \Wpo\Services\Log_Service;
use \Wpo\Services\Options_Service;

if (!class_exists('\Wpo\Sync\SyncV2_Helpers')) {

    class Sync_Helpers
    {
        /**
         * Checks if the user can retrieve an access token for the requested scope.
         * 
         * @since   24.0            Moved to Sync_Helpers
         * 
         * @param   string $scope   Scope for which the token must be valid.
         * @return  bool|WP_Error   True if user can retrieve an access token for the requested scope otherwise a WP_Error is returned.
         */
        public static function check_permissions($request)
        {
            if (!wp_verify_nonce($request->get_header('X-WP-Nonce'), 'wp_rest')) {
                return new \WP_Error('UnauthorizedException', 'The request cannot be validated.', array('status' => 401));
            }

            $wp_usr = \wp_get_current_user();

            if (empty($wp_usr)) {
                return new \WP_Error('UnauthorizedException', 'Please sign in first before using this API.', array('status' => 401));
            }

            if (!Permissions_Helpers::user_is_admin($wp_usr)) {
                return new \WP_Error('UnauthorizedException', 'Please sign in with administrative credentials before using this API.', array('status' => 403));
            }

            return true;
        }

        /**
         * Will try and retrieve a user sync job by ID from the WPO365 configuration.
         * 
         * @since   15.0
         * @since   24.0            Moved to Sync_Helpers
         * 
         * @param   string $job_id  The ID of the job.
         * @return  array|WP_Error  he job found or WP_Error if not found.
         */
        public static function get_user_sync_job_by_id($job_id)
        {
            $jobs = Options_Service::get_global_list_var('user_sync_jobs', false);
            $job = 0;

            foreach ($jobs as $_job) {

                if ($_job['id'] == $job_id) {
                    $job = $_job;
                    break;
                }
            }

            if (empty($job)) {
                return new WP_Error('NotFoundException', sprintf('%s -> User synchronization job with ID %s not found.', __METHOD__, $job_id));
            }

            return $job;
        }

        /**
         * @since 10.0
         * 
         * @param $delete bool Whether or not the scheduled events should be deleted
         * @return array Collection of scheduled events
         */
        public static function get_scheduled_events($job_id, $delete = false, $wp_to_aad = false)
        {
            $cron_jobs = _get_cron_array();
            $wpo_sync_jobs = array();

            foreach ($cron_jobs as $timestamp => $array_of_jobs) {

                foreach ($array_of_jobs as $hook => $jobs) {

                    if (!$wp_to_aad && ($hook == 'wpo_sync_users_start' || $hook == 'wpo_sync_users' || $hook == 'wpo_sync_v2_users_start' || $hook == 'wpo_sync_v2_users_next')) {

                        foreach ($jobs as $id => $job) {
                            $job['hook'] = $hook; // Add the hook back so it can be used.
                            $wpo_sync_jobs[$timestamp] = $job;

                            if ($delete) {

                                if (empty($job_id) || $job['args'][0] == $job_id) {
                                    $nr_of_unscheduled_events = wp_clear_scheduled_hook($hook, $job['args']);
                                    Log_Service::write_log('DEBUG', __METHOD__ . ' -> Unscheduled ' . $nr_of_unscheduled_events . ' cron jobs [hook: ' . $hook . ']');
                                }
                            }
                        }
                    }

                    if ($wp_to_aad && ($hook == 'wpo_sync_wp_to_aad_start' || $hook == 'wpo_sync_wp_to_aad_next')) {

                        foreach ($jobs as $id => $job) {
                            $job['hook'] = $hook; // Add the hook back so it can be used.
                            $wpo_sync_jobs[$timestamp] = $job;

                            if ($delete) {

                                if (empty($job_id) || $job['args'][0] == $job_id) {
                                    $nr_of_unscheduled_events = wp_clear_scheduled_hook($hook, $job['args']);
                                    Log_Service::write_log('DEBUG', __METHOD__ . ' -> Unscheduled ' . $nr_of_unscheduled_events . ' cron jobs [hook: ' . $hook . ']');
                                }
                            }
                        }
                    }
                }
            }

            return $wpo_sync_jobs;
        }

        /**
         * Updates the job provided in the array of jobs that are stored 
         * as part of the WPO365 settings.
         * 
         * @since 15.0
         * 
         * @param   array   $job        The job to be updated in the array
         * @return  boolean|WP_Error    True if no error occurred otherwise WP_Error
         */
        public static function update_user_sync_job($job)
        {
            $jobs = Options_Service::get_global_list_var('user_sync_jobs', false);
            $job_index = -1;

            foreach ($jobs as $i => $_job) {

                if ($_job['id'] == $job['id']) {
                    $job_index = $i;
                    break;
                }
            }

            if ($job_index == -1) {
                return new \WP_Error('NotFoundException', __METHOD__ . ' -> User synchronization stopped [Job with ID ' . $job['id'] . ' not found.]');
            }

            $jobs[$job_index] = $job;
            Options_Service::add_update_option('user_sync_jobs', $jobs);

            return true;
        }

        /**
         * Action handler that will check every minute whether there is a stale WPO365 User Synchronization Job (a job is
         * considered stale if it was last updated 5 minutes ago and no WP Cron Job to execute the next batch has been created). 
         * 
         * @since 28.x
         * 
         * @param mixed $job_id 
         * @return void 
         */
        public static function user_sync_monitor($job_id)
        {
            // Get latest version of $job by its id
            $jobs = Options_Service::get_global_list_var('user_sync_jobs', false);
            $job = false;

            foreach ($jobs as $_job) {

                if ($_job['id'] == $job_id) {
                    $job = $_job;
                    break;
                }
            }

            // Bail out if job was not found
            if (empty($job)) {
                Log_Service::write_log('WARN', sprintf('%s -> Cannot monitor WPO365 User Sync job with ID %s [job empty]', __METHOD__, $job_id));
                return;
            }

            // Bail out if job.last is not an array
            if (!is_array($job['last'])) {
                Log_Service::write_log('WARN', sprintf('%s -> Cannot monitor WPO365 User Sync job with ID %s [job last missing]', __METHOD__, $job_id));
                return;
            }

            // Unschedule monitor once job.last.stopped is true
            if (isset($job['last']['stopped']) && true === $job['last']['stopped']) {
                $result = wp_clear_scheduled_hook('wpo_sync_v2_monitor', array($job_id), true);
                return;
            }

            /**
             * The monitor will run every minute but only after the WPO365 User Sync Job has not been 
             * updated since 300 seconds will it check if the event is still scheduled.
             */

            if (time() - $job['last']['date'] >= 300) {
                $cron_jobs = _get_cron_array();

                foreach ($cron_jobs as $timestamp => $_array_of_jobs) {

                    foreach ($_array_of_jobs as $_hook => $_jobs) {

                        if ($_hook == 'wpo_sync_v2_users_next') {

                            foreach ($_jobs as $_id => $_job) {

                                if ($job['id'] == $_job['args'][0]) {
                                    $next_job_found = true;
                                    Log_Service::write_log('DEBUG', sprintf('%s -> Event "wpo_sync_v2_users_next" for WPO365 User Sync Job with ID %s scheduled [nothing to do]', __METHOD__, $job['id']));
                                }
                            }
                        }
                    }
                }

                if (empty($next_job_found)) {
                    Log_Service::write_log('ERROR', sprintf('%s -> Event "wpo_sync_v2_users_next" for WPO365 User Sync Job with ID %s not scheduled [event will be rescheduled]', __METHOD__, $job['id']));
                    $random_arg = uniqid();
                    wp_schedule_single_event(time(), 'wpo_sync_v2_users_next', [$job['id'], null, $random_arg]);
                }
            }
        }

        /**
         * Adds a new WP Cron Job that will check every minute to see if there is a stale
         * WPO365 User Synchronization Job.
         * 
         * @since 28.x
         * 
         * @param mixed $job 
         * @return void 
         */

        public static function schedule_user_sync_monitor($job)
        {
            if (empty($job) || empty($job['id'])) {
                Log_Service::write_log('WARN', sprintf('%s -> Cannot add WPO365 User Sync Monitor [job is empty or id is missing]', __METHOD__));
                return;
            }

            if (false === wp_next_scheduled('wpo_sync_v2_monitor', array($job['id']))) {
                Log_Service::write_log('DEBUG', sprintf('%s -> Adding WPO365 User Sync Monitor', __METHOD__));
                wp_schedule_event(time(), 'wpo_every_minute', 'wpo_sync_v2_monitor', array($job['id']));
                return;
            }

            Log_Service::write_log('DEBUG', sprintf('%s -> Cannot add WPO365 User Sync Monitor [job already added]', __METHOD__));
        }

        /**
         * 
         * @since 28.x  Restores for each user sync job its "last" cache from the database.
         * 
         * @param mixed $options 
         * @param mixed $jobs 
         * @return void 
         */
        public static function restore_jobs_last(&$options, $jobs)
        {
            if (is_array($options['user_sync_jobs'])) {

                foreach ($jobs as $job) {

                    if (empty($job['last'])) {
                        continue;
                    }

                    for ($i = 0; $i < count($options['user_sync_jobs']); $i++) {

                        if ($options['user_sync_jobs'][$i]['id'] == $job['id']) {
                            $options['user_sync_jobs'][$i]['last'] = $job['last'];
                        }
                    }
                }
            }
        }
    }
}
